Schema = require("../index").Schema
Text = Schema.Text

require("./spec_helper").init exports

schemas =
  mongoose:
    database: "senseorm_test"
  mysql:
    database: "senseorm_test"
    username: "root"
  redis: {}
  memory: {}
  riak: {}

testOrm = (schema) ->
  allTestsDone = ->
    schema.disconnect()
    console.log "Test done in %dms\n", Date.now() - start
  Post = User =  undefined
  start = Date.now()
  it "should define class", (test) ->
    User = schema.define("User",
      name: String
      bio: Text
      approved: Boolean
      joinedAt: Date
      age: Number
    )
    Post = schema.define("Post",
      title:
        type: String
        length: 255
        index: true

      content:
        type: Text

      date:
        type: Date
        default: Date.now

      published:
        type: Boolean
        default: false
    )
    Post.validateAsync "title", (err, done) ->
      process.nextTick done

    User.hasMany Post,
      as: "posts"
      foreignKey: "user_id"

    Post.belongsTo User,
      as: "author"
      foreignKey: "user_id"

    user = new User
    test.ok User instanceof Function
    test.ok User.find instanceof Function
    test.ok User.create instanceof Function
    test.ok user.save instanceof Function
    schema.automigrate (err) ->
      if err
        console.log "Error while migrating"
        console.log err
      else
        test.done()

  it "should initialize object properly", (test) ->
    hw = "Hello word"
    now = Date.now()
    post = new Post(title: hw)
    test.equal post.title, hw
    test.ok not post.propertyChanged("title")
    post.title = "Goodbye, Lenin"
    test.equal post.title_was, hw
    test.ok post.propertyChanged("title")
    test.strictEqual post.published, false
    test.ok post.date >= now
    test.ok post.isNewRecord()
    test.done()

  it "should be expoted to JSON", (test) ->
    test.equal JSON.stringify(new Post(
      id: 1
      title: "hello, json"
      date: 1
    )), "{\"id\":1,\"title\":\"hello, json\",\"content\":null,\"date\":1,\"published\":false,\"user_id\":null}"
    test.done()

  it "should create object", (test) ->
    Post.create (err, post) ->
      throw err  if err
      test.ok post.id
      test.ok not post.title, "Title is blank"
      Post.exists post.id, (err, exists) ->
        throw err  if err
        test.ok exists
        test.done()

  it "should save object", (test) ->
    title = "Initial title"
    title2 = "Hello world"
    date = new Date
    Post.create
      title: title
      date: date
    , (err, obj) ->
      test.ok obj.id, "Object id should present"
      test.equals obj.title, title
      obj.title = title2
      test.ok obj.propertyChanged("title"), "Title changed"
      obj.save (err, obj) ->
        test.equal obj.title, title2
        test.ok not obj.propertyChanged("title")
        test.done()

  it "should create object with initial data", (test) ->
    title = "Initial title"
    date = new Date
    Post.create
      title: title
      date: date
    , (err, obj) ->
      test.ok obj.id
      test.equals obj.title, title
      test.equals obj.date, date
      Post.find obj.id, ->
        test.equal obj.title, title
        test.equal obj.date.toString(), date.toString()
        test.done()

  it "should not create new instances for the same object", (test) ->
    title = "Initial title"
    Post.create
      title: title
    , (err, post) ->
      test.ok post.id, "Object should have id"
      test.equals post.title, title
      Post.find post.id, (err, foundPost) ->
        throw err  if err
        test.equal post.title, title
        test.strictEqual post, foundPost
        test.done()

  it "should not re-instantiate object on saving", (test) ->
    title = "Initial title"
    post = new Post(title: title)
    post.save (err, savedPost) ->
      test.strictEqual post, savedPost
      test.done()

  it "should destroy object", (test) ->
    Post.create (err, post) ->
      Post.exists post.id, (err, exists) ->
        test.ok exists, "Object exists"
        post.destroy ->
          Post.exists post.id, (err, exists) ->
            console.log err  if err
            test.ok not exists, "Hey! ORM told me that object exists, but it looks like it doesn't. Something went wrong..."
            Post.find post.id, (err, obj) ->
              test.equal obj, null, "Param obj should be null"
              test.done()

  it "should update single attribute", (test) ->
    Post.create
      title: "title"
      content: "content"
      published: true
    , (err, post) ->
      post.content = "New content"
      post.updateAttribute "title", "New title", ->
        test.equal post.title, "New title"
        test.ok not post.propertyChanged("title")
        test.equal post.content, "New content"
        test.ok post.propertyChanged("content")
        post.reload ->
          test.equal post.title, "New title"
          test.ok not post.propertyChanged("title")
          test.equal post.content, "content"
          test.ok not post.propertyChanged("content")
          test.done()

  countOfposts = undefined
  it "should fetch collection", (test) ->
    Post.all (err, posts) ->
      countOfposts = posts.length
      test.ok countOfposts > 0
      test.ok posts[0] instanceof Post
      test.done()

  it "should fetch count of records in collection", (test) ->
    Post.count (err, count) ->
      test.equal countOfposts, count
      test.done()

  it "should find filtered set of records", (test) ->
    done = ->
      test.done()  if --wait is 0
    wait = 3
    Post.all
      where:
        title: "New title"
    , (err, res) ->
      pass = true
      res.forEach (r) ->
        pass = false  unless r.title is "New title"

      test.ok res.length > 0
      test.ok pass, "Exact match with string"
      done()

    Post.all
      where:
        title: null
    , (err, res) ->
      pass = true
      res.forEach (r) ->
        pass = false  if r.title?

      test.ok res.length > 0
      test.ok pass, "Matching null"
      done()

    Post.all
      where:
        title: /hello/i
    , (err, res) ->
      pass = true
      res.forEach (r) ->
        pass = false  if not r.title or not r.title.match(/hello/i)

      test.ok res.length > 0
      test.ok pass, "Matching regexp"
      done()

  it "should handle hasMany relationship", (test) ->
    User.create (err, u) ->
      return console.log(err)  if err
      test.ok u.posts, "Method defined: posts"
      test.ok u.posts.build, "Method defined: posts.build"
      test.ok u.posts.create, "Method defined: posts.create"
      u.posts.create (err, post) ->
        return console.log(err)  if err
        test.ok post.author(), u.id
        u.posts (err, posts) ->
          test.strictEqual posts.pop(), post
          test.done()

  it "should support scopes", (test) ->
    done = ->
      test.done()  if --wait is 0
    wait = 2
    test.ok Post.scope, "Scope supported"
    Post.scope "published",
      where:
        published: true

    test.ok typeof Post.published is "function"
    test.ok Post.published._scope.published = true
    post = Post.published.build()
    test.ok post.published, "Can build"
    test.ok post.isNewRecord()
    Post.published.create (err, psto) ->
      return console.log(err)  if err
      test.ok psto.published
      test.ok not psto.isNewRecord()
      done()

    User.create (err, u) ->
      return console.log(err)  if err
      test.ok typeof u.posts.published is "function"
      test.ok u.posts.published._scope.where.published
      test.equal u.posts.published._scope.where.user_id, u.id
      done()

  it "should destroy all records", (test) ->
    Post.destroyAll (err) ->
      if err
        console.log "Error in destroyAll"
        throw err
      Post.all (err, posts) ->
        test.equal posts.length, 0
        Post.count (err, count) ->
          test.equal count, 0
          test.done()
          process.nextTick allTestsDone


Object.keys(schemas).forEach (schemaName) ->
  return if process.env.ONLY && process.env.ONLY != schemaName
  context schemaName, ->
    schema = new Schema schemaName, schemas[schemaName]
    testOrm(schema)